/*
 * Copyright (C) 2005-2014 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file recv.h Receiver-related structures and exported symbols of recv_common.c. */

/**
 * Describes one instance of a receiver.
 */
struct receiver_node {
	/** Points to the corresponding receiver. */
	struct receiver *receiver;
	/** Receiver-specific data. */
	void *private_data;
	/** Pointer to the configuration data for this instance. */
	void *conf;
	/** The task associated with this instance. */
	struct task *task;
	/** The receiver node is always the root of the buffer tree. */
	struct btr_node *btrn;
	/** Each receiver node maintains a buffer pool for the received data. */
	struct btr_pool *btrp;
	/**
	 * The file descriptor to receive the stream.
	 *
	 * The pre_select function of the receiver adds this file descriptor to
	 * the set of file descriptors which are watched for readability or
	 * writability, depending on the state of the connection (if any).
	 *
	 * If \a fd is readable, the post_select function of the receiver reads
	 * data from this fd into the buffer pool area of \a btrp.
	 *
	 * \sa \ref receiver.
	 */
	int fd;
};

/**
 * Describes one supported paraslash receiver.
 *
 * \sa http_recv.c, udp_recv.c
 */
struct receiver {
	/**
	 * The name of the receiver.
	 */
	const char *name;
	/**
	 * The receiver init function.
	 *
	 * It must fill in all other function pointers and is assumed to succeed.
	 *
	 * \sa http_recv_init udp_recv_init.
	 */
	void (*init)(struct receiver *r);
	/**
	 * The command line parser of the receiver.
	 *
	 * It should check whether the command line options given by \a argc
	 * and \a argv are valid.  On success, it should return a pointer to
	 * the receiver-specific configuration data determined by \a argc and
	 * \a argv.  Note that this might be called more than once with
	 * different values of \a argc and \a argv.
	 */
	void *(*parse_config)(int argc, char **argv);
	/**
	 * Deallocate the configuration structure of a receiver node.
	 *
	 * This calls the receiver-specific cleanup function generated by
	 * gengetopt.
	 */
	void (*free_config)(void *conf);
	/**
	 * Open one instance of the receiver.
	 *
	 * This should allocate the output buffer of \a rn. and may also
	 * perform any other work necessary for retrieving the stream according
	 * to the configuration stored in the \a conf member of \a rn which is
	 * guaranteed to point to valid configuration data (as previously
	 * obtained from the config parser).
	 *
	 * \sa receiver_node::conf, receiver_node::buf.
	 */
	int (*open)(struct receiver_node *rn);
	/**
	 * Close this instance of the receiver.
	 *
	 * It should free all resources associated with given receiver node
	 * that were allocated during the corresponding open call.
	 *
	 * \sa receiver_node.
	 */
	void (*close)(struct receiver_node *rn);
	/**
	 * Add file descriptors to fd_sets and compute timeout for select(2).
	 *
	 * The pre_select function gets called from the driving application
	 * before entering its select loop. The receiver may use this hook to
	 * add any file descriptors to the sets of file descriptors given by \a
	 * s.
	 *
	 * \sa select(2), time.c struct task, struct sched.
	 */
	void (*pre_select)(struct sched *s, void *context);
	/**
	 * Evaluate the result from select().
	 *
	 * This hook gets called after the call to select(). It should check
	 * all file descriptors which were added to any of the fd sets during
	 * the previous call to pre_select. According to the result, it may
	 * then use any non-blocking I/O to establish a connection or to
	 * receive the audio data.
	 *
	 * \sa select(2), struct receiver.
	 */
	int (*post_select)(struct sched *s, void *context);

	/** The two help texts of this receiver. */
	struct ggo_help help;
	/**
	 * Answer a buffer tree query.
	 *
	 * This optional function pointer is used for inter node communications
	 * of the buffer tree nodes. See \ref btr_command_handler for details.
	 */
	btr_command_handler execute;
};

/** Define an array of all available receivers. */
#define DEFINE_RECEIVER_ARRAY struct receiver receivers[] = { \
	HTTP_RECEIVER \
	DCCP_RECEIVER \
	UDP_RECEIVER \
	AFH_RECEIVER \
	{.name = NULL}};

/** Iterate over all available receivers. */
#define FOR_EACH_RECEIVER(i) for (i = 0; receivers[i].name; i++)

void recv_init(void);
void *check_receiver_arg(char *ra, int *receiver_num);
void print_receiver_helps(unsigned flags);
int generic_recv_pre_select(struct sched *s, struct receiver_node *rn);

/** \cond receiver */
extern void http_recv_init(struct receiver *r);
#define HTTP_RECEIVER {.name = "http", .init = http_recv_init},
extern void dccp_recv_init(struct receiver *r);
#define DCCP_RECEIVER {.name = "dccp", .init = dccp_recv_init},
extern void udp_recv_init(struct receiver *r);
#define UDP_RECEIVER {.name = "udp", .init = udp_recv_init},
#define AFH_RECEIVER /* not active by default */

extern struct receiver receivers[];
/** \endcond receiver */

