/*
 * Copyright (C) 2004-2014 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */
/** \file signal.c Signal handling functions. */

#include <signal.h>
#include <sys/types.h>

#include "para.h"
#include "error.h"
#include "fd.h"
#include "list.h"
#include "sched.h"
#include "signal.h"

static int signal_pipe[2];

/**
 * Initialize the paraslash signal subsystem.
 *
 * This function creates a pipe, the signal pipe, to deliver pending
 * signals to the application (Bernstein's trick). It should be called
 * during the application's startup part, followed by subsequent calls
 * to para_install_sighandler() for each signal that should be caught.
 *
 * A generic signal handler is used for all signals simultaneously. When a
 * signal arrives, the signal handler writes the number of the signal received
 * to one end of the signal pipe. The application can test for pending signals
 * by checking if the file descriptor of the other end of the signal pipe is
 * ready for reading, see select(2).
 *
 * \return This function either succeeds or calls exit(2) to terminate the
 * current process. On success, the file descriptor of the read end of the
 * signal pipe is returned.
 */
int para_signal_init(void)
{
	int ret;
	if (pipe(signal_pipe) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err_out;
	}
	ret = mark_fd_nonblocking(signal_pipe[0]);
	if (ret < 0)
		goto err_out;
	ret = mark_fd_nonblocking(signal_pipe[1]);
	if (ret < 0)
		goto err_out;
	return signal_pipe[0];
err_out:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

/* Write the signal number to signal pipe. */
static void generic_signal_handler(int s)
{
	/*
	 * Signal handlers that make system calls must save a copy of errno on
	 * entry to the handler and restore it on exit, to prevent the
	 * possibility of overwriting a errno value that had previously been
	 * set in the main program.
	 */
	int save_errno = errno;
	ssize_t ret = write(signal_pipe[1], &s, sizeof(int));

	if (ret == sizeof(int)) {
		errno = save_errno;
		return;
	}
	if (ret < 0)
		PARA_EMERG_LOG("%s\n", strerror(errno));
	else
		PARA_EMERG_LOG("short write to signal pipe\n");
	exit(EXIT_FAILURE);
}

/**
 * Reap one child.
 *
 * \param pid In case a child died, its pid is returned here.
 *
 * Call waitpid() and print a log message containing the pid and the cause of
 * the child's death.
 *
 * \return A (negative) paraslash error code on errors, zero, if no child died,
 * one otherwise. If and only if the function returns one, the content of \a
 * pid is meaningful.
 *
 * \sa waitpid(2).
 */
int para_reap_child(pid_t *pid)
{
	int status;
	*pid = waitpid(-1, &status, WNOHANG);

	if (!*pid)
		return 0;
	if (*pid < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	if (WIFEXITED(status))
		PARA_DEBUG_LOG("child %i exited. Exit status: %i\n", (int)*pid,
			WEXITSTATUS(status));
	else if (WIFSIGNALED(status))
		PARA_DEBUG_LOG("child %i was killed by signal %i\n", (int)*pid,
			WTERMSIG(status));
	else
		PARA_WARNING_LOG("child %i terminated abormally\n", (int)*pid);
	return 1;
}

/**
 * Install the given handler for the given signal.
 *
 * \param sig The number of the signal to catch.
 * \param handler to be installed, \p SIG_IGN, or \p SIG_DFL.
 *
 * This either succeeds or calls exit(EXIT_FAILURE).
 *
 * \sa sigaction(2).
 */
void para_sigaction(int sig, void (*handler)(int))
{
	struct sigaction act;

	PARA_DEBUG_LOG("catching signal %d\n", sig);
	act.sa_handler = handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if (sig == SIGALRM) {
		#ifdef SA_INTERRUPT /* SunOS */
			act.sa_flags |= SA_INTERRUPT;
		#endif
	} else {
		#ifdef SA_RESTART /* BSD */
			act.sa_flags |= SA_RESTART;
		#endif
	}
	if (sigaction(sig, &act, NULL) >= 0)
		return;
	PARA_EMERG_LOG("failed to install signal handler for signal %d\n",
		sig);
	exit(EXIT_FAILURE);
}

/**
 * Install the generic signal handler for the given signal number.
 *
 * \param sig The number of the signal to catch.
 *
 * \sa signal(2), sigaction(2).
 */
void para_install_sighandler(int sig)
{
	para_sigaction(sig, &generic_signal_handler);
}

/**
 * Block a signal for the caller.
 *
 * \param sig The signal to block.
 *
 * This sets the given signal in the current signal mask of the calling process
 * to prevent this signal from delivery.
 *
 * \sa \ref para_unblock_signal(), sigprocmask(2), sigaddset(3).
 */
void para_block_signal(int sig)
{
	sigset_t set;

	PARA_DEBUG_LOG("blocking signal %d\n", sig);
	sigemptyset(&set);
	sigaddset(&set, sig);
	sigprocmask(SIG_BLOCK, &set, NULL);
}

/**
 * Unblock a signal.
 *
 * \param sig The signal to unblock.
 *
 * This function removes the given signal from the current set of blocked
 * signals.
 *
 * \sa \ref para_block_signal(), sigprocmask(2), sigaddset(3).
 */
void para_unblock_signal(int sig)
{
	sigset_t set;

	PARA_DEBUG_LOG("unblocking signal %d\n", sig);
	sigemptyset(&set);
	sigaddset(&set, sig);
	sigprocmask(SIG_UNBLOCK, &set, NULL);
}

/**
 * Return the number of the next pending signal.
 *
 * \param rfds The fd_set containing the signal pipe.
 *
 * \return On success, the number of the received signal is returned. If there
 * is no signal currently pending, the function returns zero. On read errors
 * from the signal pipe, the process is terminated.
 */
int para_next_signal(fd_set *rfds)
{
	size_t n;
	int s, ret = read_nonblock(signal_pipe[0], &s, sizeof(s), rfds, &n);

	if (ret < 0) {
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	if (n == 0)
		return 0;
	assert(n == sizeof(s));
	PARA_DEBUG_LOG("next signal: %d\n", s);
	return s;
}

/**
 * Close the write end of the signal pipe.
 */
void para_signal_shutdown(void)
{
	close(signal_pipe[1]);
}
