/*
 * Copyright (C) 2013-2014 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file version.c Some helpers for printing version and copyright strings. */

#include "para.h"
#include "version.h"
#include "git-version.h"

/**
 * Get the raw git version string
 *
 * \return The string generated by the GIT-VERSION-GEN script. It is passed
 * as a preprocessor define during compilation.
 */
__a_const const char *version_git(void)
{
	return GIT_VERSION;
}

/**
 * Get the version string for an executable.
 *
 * \param pfx The program name (without the leading "para_").
 *
 * \return A statically allocated string which contains the program name and
 * the git version. It must not be freed by the caller.
 */
const char *version_single_line(const char *pfx)
{
	static char buf[100];
	snprintf(buf, sizeof(buf) - 1,
		"para_%s " GIT_VERSION, pfx);
	return buf;
}

/**
 * Get the full version text.
 *
 * \param pfx See \ref version_single_line().
 *
 * \return A string containing the same text as returned by \ref
 * version_single_line(), augmented by additional build information, a
 * copyright text and the email address of the author.
 *
 * Like \ref version_single_line(), this string is stored in a statically
 * allocated buffer and must not be freed.
 */
const char *version_text(const char *pfx)
{
	static char buf[512];

	snprintf(buf, sizeof(buf) - 1, "%s\n"
		"Copyright (C) 2014 Andre Noll\n"
		"This is free software with ABSOLUTELY NO WARRANTY."
		" See COPYING for details.\n"
		"Report bugs to <maan@tuebingen.mpg.de>.\n"
		"build date: " BUILD_DATE ",\n"
		"build system: " UNAME_RS ",\n"
		"compiler: " CC_VERSION ".\n",
		version_single_line(pfx)
	);
	return buf;
}

/**
 * Print the version text and exit successfully.
 *
 * \param pfx See \ref version_single_line().
 * \param flag Whether --version was given.
 *
 * If \a flag is false, this function does nothing. Otherwise it prints the
 * full version text as returned by \ref version_text() and exits successfully.
 */
void version_handle_flag(const char *pfx, bool flag)
{
	if (!flag)
		return;
	printf("%s", version_text(pfx));
	exit(EXIT_SUCCESS);
}
